/* ####################################################################
 * jPVM - an interface to the Parallel Virtual Machine for Java
 * ####################################################################
 *
 * --------------------------------------------------------------------
 * $Id: pvmInterface.c,v 1.23 1998/04/10 17:42:18 dave Exp $
 * --------------------------------------------------------------------
 *
 * Copyright (c) 1996 by:
 *
 *		David A.Thurman
 *		Center for Human-Machine Systems Research
 *		School of Industrial and Systems Engineering
 *		Georgia Institute of Technology
 *		Atlanta, Georgia, USA
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * --------------------------------------------------------------------
 *
 * $Log: pvmInterface.c,v $
 * Revision 1.23  1998/04/10 17:42:18  dave
 * Name change to jPVM.
 * Fixed bug in pvm_spawn().
 *
 * Revision 1.22  1998/01/19 16:19:31  dave
 * Changes for v1.1.3b distribution.
 *
 * Revision 1.20  1997/09/20 00:05:24  dave
 * minor changes.
 *
 * Revision 1.19  1997/09/17 11:41:22  dave
 * Removed pvm_ prefix on all methods, since they will now be
 * referenced as jPVM.method()
 *
 * Revision 1.20  1997/09/12 20:13:12  dave
 * Fixed bugs in pvm_upk* & pvm_pk* routines involving getting and
 * releasing Java arrays.
 * Cleaned up variable names for readability and consistency.
 * Modified methods which no longer needed array length arguments.
 * Removed pvm_ prefix to methods
 *
 * Revision 1.19  1997/09/12 15:28:54  dave
 * Added pvm_trecv()
 * Modified pvm_addhosts() and pvm_delhosts(): Removed nhosts argument
 * since Java provides length of array.  Need to do same in other routines.
 * Changed many variable names for readability and consistency.
 * Shorted w/in method documentation in places.
 *
 * Revision 1.18  1997/09/11 13:44:24  ippolito
 * Added functionality for all the functions except tfor pk complex numbers.
 *
 * Revision 1.17  1997/09/10 19:09:14  ippolito
 * Redid pk* and upk* algorithms to use proper length packing
 * and unpacking procedures for Solaris and Irix.
 *
 * Revision 1.16  1997/08/29 17:31:24  ippolito
 * Updated the upk* and pk* methods for byte, double, float, long
 * and short.
 *
 * Revision 1.15  1997/08/22 21:55:35  dave
 * Turned off debugging output for testing with benchmarking
 * application.
 *
 * Revision 1.14  1997/08/21 15:42:23  dave
 * Changed comment style from C++ to ANSI-C so people with
 * other compilers won't get messed up.
 *
 * Revision 1.13  1997/08/21 15:23:51  dave
 * Changed method names that are printed in the debugging output.
 *
 * Revision 1.12  1997/08/21 15:20:53  dave
 * Changed names of routines back to having the 'pvm_' prefix in
 * jPVM class, which meant the names got a bit more complicated
 * in here, including the addition of a '_1' in the middle of the
 * method names (see JNI documentation if you care why).
 *
 * Revision 1.11  1997/08/21 14:04:22  dave
 * Fixed bug in attempted use of macro DEBUG_MSG for debugging output.
 * Got rid of macro and returned to more standard if (_debug) style.
 *
 * Revision 1.10  1997/08/21 13:48:15  dave
 * Implemented JNI functions used in jPVM example program, including:
 * initsend, pkint, bcast, barrier, joingroup, pkstr, upkstr, upkint,
 * mytid, and exit.
 *
 * Revision 1.9  1997/08/20 18:33:47  dave
 * Beginning of changes over to JNI.  Have only changed implementation
 * of pvm_addhosts and pvm_barrier so far.
 *
 * Revision 1.8  1997/08/20 15:43:04  dave
 * Changed #include <java_lang_Integer.h> to
 * #include "java_lang_Integer.h" so that it would be found in
 * the current directory.
 *
 * Revision 1.7  1996/12/16 21:04:28  dave
 * Removed #ifdef DEBUG syntax and replace with code to check
 * _debug flag in jPVM object passed to all methods to determine
 * whether or not to spit out debugging information.
 * Modified jPVM.pvm_upkstr() to use char[] as argument in keeping
 * with changes in jPVM.
 *
 * Revision 1.6  1996/06/03 15:56:55  dave
 * Started adding pvm_config and pvm_reg_rm using new Java classes.
 * Commented out debugging information in pk_double.
 * Fixed bug in pvm_upkstr.  Routine now allocates space for
 * string to be unpacked into, then uses that space to update
 * the Java string.
 *
 * Revision 1.5  1996/05/30 14:43:44  dave
 * Renamed file from jPVM.c to pvmInterface.c
 *
 * Revision 1.4  1996/05/30 14:00:38  dave
 * Added jPVM.pvm_parent
 * Added many debugging statements
 * Added memory allocation/deallocation to routines which get arrays
 * of Strings from Java.
 * Fixed bug in upkstr to change the value of the Java String
 * w/o allocating a new String.
 *
 * Revision 1.3  1996/05/23 14:16:36  dave
 * Added pvm_reg_hoster and pvm_reg_tasker
 *
 * Revision 1.2  1996/05/23 14:03:32  dave
 * Implemented all functions.  Have not tested them, however, it all
 * compiles.
 *
 * Revision 1.1  1996/05/21 13:44:50  dave
 * Initial revision
 *
 *
 * --------------------------------------------------------------------
 */

#include <stdio.h>
#include <limits.h>

#ifdef WIN32
#include <time.h>
#else
#include <sys/time.h>
#endif

#include <pvm3.h>

#include "jPVM.h"

/* --------------------------------------------------------------------
 * Note to parameter names
 *
 * (1) 	With the exception of the 'env' and 'obj' arguments in all
 *     	methods, all parameter names in native methods should start 
 *	with 'j' followed by a capitalized word describing the 
 *	argument.  Example 'jHosts' is a list of host Strings.
 * (2)  Variables within routines where are used to make copies of
 * 	arguments should start with a 'c' followed by the same 
 *	capitalized word used in the method parameter of which they
 *	are a copy.
 * (3) 	The 'jInfo' variable used in every routine should be of type
 *	'jint'.  The result from the call to the actual PVM routine
 *	should be cast (jint) before being assigned to info.
 * (4)	Character string arguments which are created via calls to the
 *	Java VM (e.g., GetStringUTFChars) should be cast to type (char *) 
 *	though care must be taken not to make any assignments to the
 *	resulting pointer, as it represents non-changeable memory.
 *	(to reduce the number of compiler warnings).
 * --------------------------------------------------------------------
 */

/*
 * Useful macros for dealing with Java strings.
 */

#define getJavaString(j)	(*env)->GetStringUTFChars(env,(j), 0)
#define releaseJavaString(j,c) 	(*env)->ReleaseStringUTFChars(env,(j),(c))

/*
 * Debugging flag.  Turn on for now, but turn off in final release.
 */

int _debug = 0;		

/* *******************************
 * jPVM.addhosts()
 * *******************************
 */

JNIEXPORT jint JNICALL
  Java_jPVM_addhosts(
	JNIEnv	 			*env, 
	jobject				obj,
	jobjectArray 			jHosts, 
	jintArray 			jInfos)
{
  char  		**cHosts;
  jstring 		*jHostsElements;
  int 	 		*cInfos;
  jint 			*jInfosBody;

  jsize			i;
  jint			jInfo;

	/* find out how big the array is */

  int 	numHosts = (*env)->GetArrayLength(env, jHosts);
  int	numInfos = (*env)->GetArrayLength(env, jInfos);

	/* error processing if array lengths don't match jNumHosts */

  if (numHosts != numInfos)
  {
      fprintf(stderr, "jPVM_pvm_addhosts: Bad match in array lengths\n");
      fprintf(stderr, "\tlength(hosts) = %d\n", numHosts);
      fprintf(stderr, "\tlength(infos) = %d\n", numInfos);
    
    return PvmBadParam;	
  }

  if (_debug)
    fprintf(stdout, "jPVM.addhosts(*,*,[],[])\n");
  
	/* allocate temporary space for C arrays */

  cHosts = (char **) calloc(numHosts, sizeof(char *));
  cInfos = (int *) calloc(numHosts, sizeof(int));
  jHostsElements = (jstring *) calloc(numHosts, sizeof(jstring));

	/* 
	 * Copy hostnames from array of Java Strings.
	 * Keep array of jstrings around to make it easier
	 * to get rid of the UTF strings later.
	 */

  for (i=0; i< numHosts; i++)
  {
    jHostsElements[i] = (jstring) (*env)->GetObjectArrayElement(env, jHosts, i);
    cHosts[i] = (char *) getJavaString(jHostsElements[i]);
  }

	/* Actual PVM Routine */

  jInfo = (jint) pvm_addhosts((char **) cHosts, numHosts, cInfos);

  if (_debug)
    fprintf(stdout, 
	"jPVM.addhost: finished adding hosts, jInfo =%d\n", jInfo);

	/*
	 * copy results back to Java array
	 * first get the body of the array, 
	 * then copy the returned values into it.
	 */

  jInfosBody = (*env)->GetIntArrayElements(env, jInfos, 0);

  for (i=0; i<numHosts; i++)
  {
    jInfosBody[i] = cInfos[i];

    if (_debug)
    fprintf(stdout, 
	"jPVM.addhost: info for %s = %d\n", cHosts[i], cInfos[i]);
  }

	/*
	 * let VM release the infos array and the host Strings
	 * then free memory allocated in this routine
	 */
 
  (*env)->ReleaseIntArrayElements(env, jInfos, jInfosBody, 0);

  for (i=0; i<numHosts; i++)
    releaseJavaString(jHostsElements[i], cHosts[i]);

  free(cHosts);
  free(cInfos);
  free(jHosts);

  if (_debug)
    fprintf(stdout, "jPVM.addhost: freed memory\n");
	
	/* return result */

  return jInfo;
}

/* *******************************
 * jPVM.barrier()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_barrier(
	JNIEnv	 			*env,
	jobject				obj,
	jstring				jGroup,
	jint				jCount)
{
  char			*cGroup;
  jint			jInfo;

	/* Create C representation of Java string */

  cGroup = (char *) getJavaString(jGroup);

  if (_debug)
    fprintf(stdout, "jPVM.barrier(%s, %d)\n", cGroup, jCount);

	/* Actual PVM routine */

  jInfo = (jint) pvm_barrier(cGroup, jCount);

	/* Free VM allocated memory */

  releaseJavaString(jGroup, cGroup);

  return jInfo;
}

/* *******************************
 * jPVM.bcast()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_bcast(
	JNIEnv	 			*env,
	jobject				obj,
	jstring				jGroup,
	jint 				jMsgtag)
{
  char		*cGroup;
  jint		jInfo;

	/* Convert group name to C string */

  cGroup = (char *) getJavaString(jGroup);

  if (_debug)
    fprintf(stdout, "jPVM.bcast(%s, %d)\n", cGroup, jMsgtag);

	/* Actual PVM routine */

  jInfo = (jint) pvm_bcast(cGroup, (int) jMsgtag);

	/* free VM allocated memory */

  releaseJavaString(jGroup, cGroup);

  return jInfo;
}

/* *******************************
 * jPVM.bufinfo()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_bufinfo(
	JNIEnv	 			*env,
	jobject				obj,
	jint				jBufId,
	jobject				jBytes,
	jobject				jMsgtag,
	jobject				jTID)
{
  int	 	cBytes;
  int		cMsgtag;
  int		cTID;
  jint		jInfo;
  jfieldID	fid;
  jclass	clazz;

  if (_debug)
    fprintf(stdout, "jPVM.bufinfo(%d, *, *, *)\n", jBufId);

	/* Actual PVM routine */

  jInfo = (jint) pvm_bufinfo((int) jBufId, &cBytes, &cMsgtag, &cTID);

	/* return if there's an erro */

  if (jInfo < 0)	 return jInfo;

	/*
	 * copy values to Java structs if no errors
	 * we know jBytes, jMsgtag, and jTID are Integers
	 * and we need their 'value' field
	 */

  clazz = (*env)->GetObjectClass(env, jBytes);
  fid = (*env)->GetFieldID(env, clazz, "value", "I");

	/* use FieldID to set values in jBtes, jMsgtag, and jTID */

  (*env)->SetIntField(env, jBytes, fid, cBytes);
  (*env)->SetIntField(env, jMsgtag, fid, cMsgtag);
  (*env)->SetIntField(env, jTID, fid, cTID);

  return jInfo;
}

/* *******************************
 * jPVM.delhosts()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_delhosts(
	JNIEnv	 			*env,
	jobject				obj,
	jobjectArray 			jHosts,
	jintArray			jInfos)
{
  char		**cHosts;
  jstring	*jHostNames;
  jint		*jInfosBody;
  int		i;
  int		numHosts;
  int		numInfos;
  int		*cInfos;
  jint		jInfo;

	/* Get array sizes */

  numHosts = (*env)->GetArrayLength(env, jHosts);
  numInfos = (*env)->GetArrayLength(env, jInfos);

  if (numHosts != numInfos)
  {
    fprintf(stderr, "jPVM_pvm_delhosts: Bad match in array lengths\n");
    fprintf(stderr, "\tlength(hosts) = %d\n", numHosts);
    fprintf(stderr, "\tlength(infos) = %d\n", numInfos);
    return PvmBadParam;	
  }

	/* allocate C arrays */

  jHostNames = (jstring *) calloc(numHosts, sizeof(jstring));
  cHosts = (char **) calloc(numHosts, sizeof(char *) );
  cInfos = (int *) calloc(numHosts, sizeof(int));

	/* copy the hostnames into the C array */

  for (i=0; i<numHosts; i++)
  {
      jHostNames[i] = (jstring) (*env)->GetObjectArrayElement(env, jHosts, i);
      cHosts[i] = (char *) getJavaString(jHostNames[i]);
  }

	/* Call PVM function */

  jInfo = (jint) pvm_delhosts(cHosts, numHosts, cInfos);

	/* Copy cInfos back to Java array */

  jInfosBody = (*env)->GetIntArrayElements(env, jInfos, 0);

  for (i=0; i<numHosts; i++)
    jInfosBody[i] = cInfos[i];
  
	/* Release the body of the array and Java strings */

  (*env)->ReleaseIntArrayElements(env, jInfos, jInfosBody, 0);

  for (i=0; i<numHosts; i++)
      releaseJavaString(jHostNames[i], cHosts[i]);

	/* Free locally allocated memory */

  free(cHosts);
  free(cInfos);
  free(jHostNames);

  return jInfo;
}

/* *******************************
 * jPVM.exit()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_exit(
	JNIEnv	 			*env,
	jobject				obj)
{
  if (_debug)
    fprintf(stdout, "jPVM.exit()\n");

  return (jint) pvm_exit();
}

/* *******************************
 * jPVM.freebuf()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_freebuf(	
	JNIEnv	 			*env,
	jobject				obj,
	jint				jBufid)
{
  if (_debug)
    fprintf(stdout, "jPVM.freebuf(%d)\n", jBufid);

  return (jint) pvm_freebuf( (int) jBufid );
}

/* *******************************
 * jPVM.getinst()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_getinst(
	JNIEnv	 			*env,
	jobject				obj,
	jstring				jGroup,
	jint				jTid)
{
  jint	jInfo;
  char	*cString;

	/* convert group to a char* */

  cString = (char *) getJavaString(jGroup);

  if (_debug)
    fprintf(stdout, "jPVM.getinst(%s, %d)\n", cString, jTid);

	/* Call the PVM function */

  jInfo = (jint) pvm_getinst(cString, (int) jTid);

	/* Release the string */

  releaseJavaString(jGroup, cString);

  return jInfo;
}

/* *******************************
 * jPVM.getopt()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_getopt(
	JNIEnv	 			*env,
	jobject				obj,
	jint				jWhat)
{
  if (_debug)
    fprintf(stdout, "jPVM.getopt()\n");

  return (jint) pvm_getopt( (int) jWhat );
}

/* *******************************
 * jPVM.getrbuf()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_getrbuf(
	JNIEnv	 			*env,
	jobject				obj)
{
  if (_debug)
    fprintf(stdout, "jPVM.getrbuf()\n");

  return (jint) pvm_getrbuf();
}

/* *******************************
 * jPVM.getsbuf()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_getsbuf(
	JNIEnv	 			*env,
	jobject				obj)
{
  jint jInfo;

  if (_debug)
    fprintf(stdout, "jPVM.getsbuf()\n");

  jInfo = (jint) pvm_getsbuf();
  return jInfo;
}

/* *******************************
 * jPVM.gettid()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_gettid(
	JNIEnv	 			*env,
	jobject				obj,
	jstring 			jGroup,
	jint				jInstance)
{
  char	*cString;
  jint  jInfo;
  
  cString = (char *) getJavaString(jGroup);

  if (_debug)
    fprintf(stdout, "jPVM.gettid(%s, %d)\n", cString, jInstance);

  jInfo = (jint) pvm_gettid(cString, (int) jInstance);

	/* Release java memory */

  releaseJavaString(jGroup, cString);
 
  return jInfo;
}

/* *******************************
 * jPVM.gsize()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_gsize(
	JNIEnv	 			*env,
	jobject				obj,
	jstring				jGroup)
{
  char	*cString;
  jint	jInfo;

  cString = (char *) getJavaString(jGroup);

  if (_debug)
    fprintf(stdout, "jPVM.gsize(%s)\n", cString);

  jInfo = (jint) pvm_gsize(cString);

	/* Release Java String */

  releaseJavaString(jGroup, cString);

  return jInfo;
}

/* *******************************
 * jPVM.halt()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_halt(
	JNIEnv	 			*env,
	jobject				obj)
{
  if (_debug)
    fprintf(stdout, "jPVM.halt()\n");

  return (jint) pvm_halt();
}

/* *******************************
 * jPVM.initsend()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_initsend(
	JNIEnv	 			*env, 
	jobject				obj,
	jint 				jEncoding)
{
  if (_debug)
    fprintf(stdout, "jPVM.initsend(%d)\n", jEncoding);

	/* Actual PVM routine */

  return (jint) pvm_initsend((int) jEncoding);
}

/* *******************************
 * jPVM.joingroup()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_joingroup(
	JNIEnv	 			*env,
	jobject				obj,
	jstring				jGroup)
{
  jint 		jInfo;
  char		*cGroup;

	/* Translate Java string to C string */

  cGroup = (char *) getJavaString(jGroup);

  if (_debug)
    fprintf(stdout, "jPVM.joingroup(%s)\n", cGroup);

	/* Actual PVM routine */

  jInfo = (jint) pvm_joingroup(cGroup);

	/* Free VM allocated memory */

  releaseJavaString(jGroup, cGroup);

  return jInfo;
}

/* *******************************
 * jPVM.kill()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_kill(
	JNIEnv	 			*env,
	jobject				obj,
	jint				jTid)
{
  if (_debug)
    fprintf(stdout, "jPVM.kill(%d)\n", jTid);

  return (jint) pvm_kill((int) jTid);
}

/* *******************************
 * jPVM.lvgroup()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_lvgroup(
	JNIEnv	 			*env,
	jobject				obj,
	jstring				jGroup)
{
  jint		jInfo;
  char		*cGroup;

  cGroup = (char *) getJavaString(jGroup);

  if (_debug)
    fprintf(stdout, "jPVM.lvgroup(%s)\n", cGroup);

  jInfo = (jint) pvm_lvgroup(cGroup);

  releaseJavaString(jGroup, cGroup);

  return jInfo;
}

/* *******************************
 * jPVM.mcast()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_mcast(
	JNIEnv	 			*env,
	jobject				obj,
	jintArray 			jTids,
	jint				jMsgtag)
{
  jint		jInfo;
  jint		*jBody;
  jsize 	jNumTasks;

	/* get size of array */

  jNumTasks = (*env)->GetArrayLength(env, jTids);
  
	/* get the body of the integer array */

  jBody = (*env)->GetIntArrayElements(env, jTids, 0);

	/* call pvm routine */

  jInfo = (jint) pvm_mcast((int *) jBody, (int) jNumTasks, (int) jMsgtag);

	/* Release the body */

  (*env)->ReleaseIntArrayElements(env, jTids, jBody, 0);

  return jInfo;

}

/* *******************************
 * jPVM.mkbuf()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_mkbuf(
	JNIEnv	 			*env,
	jobject				obj,
	jint				jEncoding)
{
  if (_debug)
    fprintf(stdout, "jPVM.mkbuf(%d)\n", jEncoding);

  return (jint) pvm_mkbuf((int) jEncoding);
}

/* *******************************
 * jPVM.mstat()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_mstat(
	JNIEnv	 			*env,
	jobject				obj,
	jstring				host)
{
  jint		jInfo;
  char		*cString;

  cString = (char *) getJavaString(host);

  if (_debug)
    fprintf(stdout, "jPVM.mstat(%s)\n", cString);

  jInfo = (jint) pvm_mstat(cString);

  releaseJavaString(host, cString);

  return jInfo;
}

/* *******************************
 * jPVM.mytid()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_mytid(
	JNIEnv				*env,
	jobject				obj)
{
  if (_debug)
    fprintf(stdout, "jPVM.mytid()\n");

	/* Actual PVM routine */

  return (jint) pvm_mytid();
}

/* *******************************
 * jPVM.notify()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_notify(
	JNIEnv	 			*env,
	jobject				obj,
	jint				jWhat,
	jint				jMsgtag,
	jintArray 			jTids)
{
  jint		jInfo;
  jint		*jBody;
  jsize		jCount;

  if (_debug)
    fprintf(stdout, 
	"jPVM.notify(%d, %d, *)\n", jWhat, jMsgtag);

	/* Get size of array and its body */

  jCount = (*env)->GetArrayLength(env, jTids);
  jBody = (*env)->GetIntArrayElements(env, jTids, 0);
 
        /* call pvm routine */
 
  jInfo = (long) pvm_notify((int) jWhat, (int) jMsgtag, (int) jCount, 
			(int *) jBody);
 
        /* Release the body */
 
  (*env)->ReleaseIntArrayElements(env, jTids, jBody, 0);

  return jInfo;
}

/* *******************************
 * jPVM.nrecv()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_nrecv(
	JNIEnv	 			*env,
	jobject				obj,
	jint				jTid,
	jint				jMsgtag)
{
  if (_debug)
    fprintf(stdout, "jPVM.nrecv(%d, %d)\n", jTid, jMsgtag);

  return (jint) pvm_nrecv( (int) jTid, (int) jMsgtag);
}

/* *******************************
 * jPVM.parent()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_parent(
	JNIEnv	 			*env,
	jobject				obj)
{
  if (_debug)
    fprintf(stdout, "jPVM.parent()\n");

  return (jint) pvm_parent();
}

/* *******************************
 * jPVM.pkbyte()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_pkbyte(
	JNIEnv	 			*env,
	jobject				obj,
	jbyteArray			jByteArray,
	jint				jNumItems,
	jint				jStride)
{
  jbyte         *jBody;     
  jsize         jArrayLen;
  jint          jInfo;     

  if (_debug)
    fprintf(stdout, "***** jPVM.pkbyte(*, %d, %d)\n", 
			jNumItems, jStride);

        /* 
	 * Get length of array
	 * Do error checking
	 * Get pointer to body of the Java array 
	 * Call PVM routine
	 * Release Java array
	 */

  jArrayLen = (*env)->GetArrayLength(env, jByteArray);
  
  if (jArrayLen < jNumItems)
  {
    fprintf(stderr, 
     "jPVM.pkbyte() error:  Array length %d less than numItems %d\n",
	jArrayLen, jNumItems);
    return PvmBadParam;
  }

  jBody = (*env)->GetByteArrayElements(env, jByteArray, 0);

  jInfo = (jint) pvm_pkbyte(jBody, (int) jNumItems, (int) jStride);

  (*env)->ReleaseByteArrayElements(env, jByteArray, jBody, 0);

  return jInfo;
}

#if 0

/* *******************************
 * jPVM.pkcplx()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_pkcplx(
	JNIEnv	 			*env,
	jobject				obj,
	jfloatArray 			cp,
	jint				nitem,
	jint				stride)
{
  if (_debug)
    fprintf(stdout, "jPVM.pkcplx(*, %d, %d)\n", nitem, stride);

  return( (jint) pvm_pkcplx(unhand(cp)->body, (int) nitem, (int) stride));
}


/* *******************************
 * jPVM.pkdcplx()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_pkdcplx(
	JNIEnv	 			*env,
	jobject				obj,
	jlongArray 			zp,
	jint				nitem,
	jint				stride)
{
  if (_debug)
    fprintf(stdout, "jPVM.pkdcplx(*, %d, %d)\n", nitem, stride);

  return( (jint) pvm_pkdcplx(unhand(zp)->body, (int) nitem, (int) stride));
}

#endif

/* *******************************
 * jPVM.pkdouble()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_pkdouble(
	JNIEnv	 			*env,
	jobject				obj,
	jdoubleArray 			jDoubleArray,
	jint				jNumItems,
	jint				jStride)
{
  jdouble	*jBody;	
  int		jArrayLen;
  jint		jInfo;

  if (_debug)
    fprintf(stdout, "jPVM.pkdouble(*, %d, %d)\n", jNumItems, jStride);

        /* 
	 * Get length of array
	 * Do error checking
 
	 * Get pointer to body of the Java array 
	 * Call PVM routine
	 * Release Java array
	 */

  jArrayLen = (*env)->GetArrayLength(env, jDoubleArray);
  
  if (jArrayLen < jNumItems)
  {
    fprintf(stderr, 
     "jPVM.pkdouble() error:  Array length %d less than numItems %d\n",
	jArrayLen, jNumItems);
    return PvmBadParam;
  }

  jBody = (*env)->GetDoubleArrayElements(env, jDoubleArray, 0);

  jInfo = (jint) pvm_pkdouble(jBody, (int) jNumItems, (int) jStride);

  (*env)->ReleaseDoubleArrayElements(env, jDoubleArray, jBody, 0);

  return jInfo;
}

/* *******************************
 * jPVM.pkfloat()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_pkfloat(
	JNIEnv	 			*env,
	jobject				obj,
	jfloatArray			jFloatArray,
	jint				jNumItems,
	jint				jStride)
{
  jfloat         *jBody;     
  jsize         jArrayLen;
  jint          jInfo;     

  if (_debug)
    fprintf(stdout, "***** jPVM.pkfloat(*,%d,%d)\n", 
		jNumItems, jStride);

        /* 
	 * Get length of array
	 * Do error checking
	 * Get pointer to body of the Java array 
	 * Call PVM routine
	 * Release Java array
	 */

  jArrayLen = (*env)->GetArrayLength(env, jFloatArray);
  
  if (jArrayLen < jNumItems)
  {
    fprintf(stderr, 
     "jPVM.pkfloat() error:  Array length %d less than numItems %d\n",
	jArrayLen, jNumItems);
    return PvmBadParam;
  }

  jBody = (*env)->GetFloatArrayElements(env, jFloatArray, 0);

  jInfo = (jint) pvm_pkfloat(jBody, (int) jNumItems, (int) jStride);

  (*env)->ReleaseFloatArrayElements(env, jFloatArray, jBody, 0);

  return jInfo;
}

/* *******************************
 * jPVM.pkint()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_pkint(
	JNIEnv	 			*env,
	jobject				obj,
	jintArray			jIntArray,
	jint				jNumItems,
	jint				jStride)
{
  jint         *jBody;     
  jint          jInfo;     
  jsize		jArrayLen;

  if (_debug)
    fprintf(stdout, "***** jPVM.pkint(*,%d,%d)\n", 
		jNumItems, jStride);

        /* 
	 * Get length of array
	 * Do error checking
	 * Get pointer to body of the Java array 
	 * Call PVM routine
	 * Release Java array
	 */

  jArrayLen = (*env)->GetArrayLength(env, jIntArray);

  if (jArrayLen < jNumItems)
  {
    fprintf(stderr, 
     "jPVM.pkint() error:  Array length %d less than numItems %d\n",
	jArrayLen, jNumItems);
    return PvmBadParam;
  }

  jBody = (*env)->GetIntArrayElements(env, jIntArray, 0);

  jInfo = (jint) pvm_pkint((int *) jBody, (int) jNumItems, (int) jStride);

  (*env)->ReleaseIntArrayElements(env, jIntArray, jBody, 0);

  return jInfo;
}

/* *******************************
 * jPVM.pklong()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_pklong(
	JNIEnv	 			*env,
	jobject				obj,
	jlongArray			jLongArray,
	jint				jNumItems,
	jint				jStride)
{
  jlong         *jBody;     
  jsize         jArrayLen;
  jint          jInfo;     

  if (_debug)
    fprintf(stdout, "***** jPVM.pklong(*,%d,%d)\n", 
			jNumItems, jStride);

        /* Get length of array
	 * Do error checking
	 * Get pointer to body of the Java array 
	 * Call PVM routine
	 * Release Java array
	 */

  jArrayLen = (*env)->GetArrayLength(env, jLongArray);

  if (jArrayLen < jNumItems)
  {
    fprintf(stderr, 
     "jPVM.pklong() error:  Array length %d less than numItems %d\n",
	jArrayLen, jNumItems);
    return PvmBadParam;
  }

  jBody = (*env)->GetLongArrayElements(env, jLongArray, 0);

	/* need error checking on sizeof(double) vs. sizeof(jlong) */

  jInfo = (jint) pvm_pkdouble((double *) jBody, (int) jNumItems, (int) jStride);

  (*env)->ReleaseLongArrayElements(env, jLongArray, jBody, 0);

  return jInfo;
}


/* *******************************
 * jPVM.pkshort()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_pkshort(
	JNIEnv	 			*env,
	jobject				obj,
	jshortArray			jShortArray,
	jint				jNumItems,
	jint				jStride)
{
  jshort         *jBody;     
  jsize         jArrayLen;
  jint          jInfo;     

  if (_debug)
    fprintf(stdout, "***** jPVM.pkshort(*,%,%d)\n", 
			jNumItems, jStride);

        /* Get length of array
	 * Do error checking
	 * Get pointer to body of the Java array 
	 * Call PVM routine
	 * Release Java array
	 */

  jArrayLen = (*env)->GetArrayLength(env, jShortArray);

  if (jArrayLen < jNumItems)
  {
    fprintf(stderr, 
     "jPVM.pkint() error:  Array length %d less than numItems %d\n",
	jArrayLen, jNumItems);
    return PvmBadParam;
  }

  jBody = (*env)->GetShortArrayElements(env, jShortArray, 0);

  jInfo = (jint) pvm_pkshort(jBody, (int) jNumItems, (int) jStride);

  (*env)->ReleaseShortArrayElements(env, jShortArray, jBody, 0);

  return jInfo;
}


/* *******************************
 * jPVM.pkstr()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_pkstr(
	JNIEnv	 			*env,
	jobject				obj,
	jstring				javaString)
{
  char		*cString;
  jint		jInfo;

  cString = (char *) getJavaString(javaString);

  if (_debug)
    fprintf(stdout, "jPVM.pkstr(%s)\n", cString);

	/* Actual PVM routine */

  jInfo = pvm_pkstr(cString);

	/* Release VM allocated memory */

  releaseJavaString(javaString, cString);

  return jInfo;
}

/* *******************************
 * jPVM.perror()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_perror(
	JNIEnv	 			*env,
	jobject				obj,
	jstring 			msg)
{

  char		*cString;
  jint		jInfo;

  cString = (char *) getJavaString(msg);

  if (_debug)
    fprintf(stdout, "jPVM.perror(%s)\n", cString);

  jInfo = (jint) pvm_perror(cString);

  releaseJavaString(msg, cString);

  return jInfo;

}


/* *******************************
 * jPVM.precv()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_precv(
	JNIEnv	 			*env,
	jobject				obj,
	jint				jTid,
	jint				jMsgtag,
	jbyteArray			jBuf,
	jint				jLen,
	jint				jDatatype,
	jobject 			jSenderTid,
	jobject			 	jMessageTag,
	jobject			 	jMessageLength)
{
  jint		jInfo;
  jbyte		*jBody;

  int		cSenderTid, cMessageTag, cMessageLength;
  jclass	jclazz;
  jfieldID	jfid;

  if (_debug)
    fprintf(stdout, "jPVM.precv(%d, %d, *, %d, *, *, *)\n",
	jTid, jMsgtag, jDatatype);

	/* Get the body of the byte array */

  jBody = (*env)->GetByteArrayElements(env, jBuf, 0);

	/* Call the actual C routine */

  jInfo = (jint) pvm_precv( (int) jTid, (int) jMsgtag, jBody, 
	(int) jLen, (int) jDatatype, 
	&cSenderTid, &cMessageTag, &cMessageLength);

	/* All the jobjects are Integer classes- fill in value field */

  jclazz = (*env)->GetObjectClass(env, jSenderTid);
  jfid = (*env)->GetFieldID(env, jclazz, "value", "I");

	/* use the FieldID to set the values of our Integer classes */

  (*env)->SetIntField(env, jSenderTid, jfid, cSenderTid);
  (*env)->SetIntField(env, jMessageTag, jfid, cMessageTag);
  (*env)->SetIntField(env, jMessageLength, jfid, cMessageLength);

	/* release the java byte array from the VM */

  (*env)->ReleaseByteArrayElements(env, jBuf, jBody, 0);
  
  return(jInfo);
}

/* *******************************
 * jPVM.probe()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_probe(
	JNIEnv	 			*env,
	jobject				obj,
	jint				tid,
	jint				msgtag)
{
  if (_debug)
    fprintf(stdout, "jPVM.probe(%d, %d)\n", tid, msgtag);

  return (jint) pvm_probe( (int) tid, (int) msgtag);
}


/* *******************************
 * jPVM.psend()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_psend(
	JNIEnv	 			*env,
	jobject				obj,
	jint				jTid,
	jint				jMsgtag,
	jbyteArray			jBuf,
	jint				jLen,
	jint				jDatatype)
{

  jint		jInfo;
  jbyte		*jBody;

  if (_debug)
    fprintf(stdout, "jPVM.psend(%d, %d, *, %d)\n",
		jTid, jMsgtag, jDatatype);

	/* Get sizeof array and array body */

  jBody = (*env)->GetByteArrayElements(env, jBuf, 0);

	/* call the actual PVM routine */

  jInfo = (jint) pvm_psend( (int) jTid, (int) jMsgtag, jBody,
                             (int) jLen, (int) jDatatype);

        /* release the java byte array from the VM */

  (*env)->ReleaseByteArrayElements(env, jBuf, jBody, 0);

  return jInfo;
  
}

/* *******************************
 * jPVM.pstat()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_pstat(
	JNIEnv	 			*env,
	jobject				obj,
	jint				tid)
{
  if (_debug)
    fprintf(stdout, "jPVM.pstat(%d)\n", tid);
  
  return (jint) pvm_pstat( (int) tid);
}

/* *******************************
 * jPVM.recv()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_recv(
	JNIEnv	 			*env, 
	jobject				obj,
	jint 				jTID, 
	jint 				jMsgtag)
{
  if (_debug)
    fprintf(stdout, "jPVM.recv(%d, %d)\n", jTID, jMsgtag);

	/* Actual PVM routine */

  return (jint) pvm_recv((int) jTID, (int) jMsgtag);
}

/* *******************************
 * jPVM.reg_hoster()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_reg_hoster(
	JNIEnv	 			*env, 
	jobject				obj)
{
  if (_debug)
    fprintf(stdout, "jPVM.reg_hoster()\n");

  return pvm_reg_hoster();
}

/* *******************************
 * jPVM.reg_tasker()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_reg_tasker(
	JNIEnv	 			*env, 
	jobject				obj)
{
  if (_debug)
    fprintf(stdout, "jPVM.reg_tasker()\n");

  return (jint) pvm_reg_tasker();
}


/* *******************************
 * jPVM.send()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_send(
	JNIEnv	 			*env, 
	jobject				obj,
	jint 				jTID, 
	jint 				jMsgtag)
{
  if (_debug)
    fprintf(stdout, "jPVM.send(%d, %d)\n", jTID, jMsgtag);

  return ( (jint) pvm_send( (int) jTID, (int) jMsgtag));
}

/* *******************************
 * jPVM.sendsig()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_sendsig(
	JNIEnv	 			*env,
	jobject				obj,
	jint				tid,
	jint				signum)
{
  if (_debug)
    fprintf(stdout, "jPVM.sendsig(%d, %d)\n", tid, signum);


  return (jint) pvm_sendsig( (int) tid, (int) signum);
}

/* *******************************
 * jPVM.setopt()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_setopt(
	JNIEnv	 			*env,
	jobject				obj,
	jint				what,
	jint				val)
{
  if (_debug)
    fprintf(stdout, "jPVM.setopt(%d, %d)\n", what, val);

  return (jint) pvm_setopt((int) what, (int) val);
}

/* *******************************
 * jPVM.setrbuf()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_setrbuf(
	JNIEnv	 			*env,
	jobject				obj,
	jint				bufid)
{
  if (_debug)
    fprintf(stdout, "jPVM.setrbuf(%d)\n", bufid);

  return (jint) pvm_setrbuf((int) bufid);
}

/* *******************************
 * jPVM.setsbuf()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_setsbuf(
	JNIEnv	 			*env,
	jobject				obj,
	jint				bufid)
{
  if (_debug)
    fprintf(stdout, "jPVM.setsbuf(%d)\n", bufid);

  return (jint) pvm_setsbuf((int) bufid);
}

/* *******************************
 * jPVM.spawn()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_spawn(
	JNIEnv	 			*env,
	jobject				obj,
	jstring				jTask,
	jobjectArray 			jArgv,
	jint				jFlag,
	jstring				jWhere,
	jintArray 			jTids)
{
  char		*cTask;
  char		*cWhere;
  jint		jInfo;
  jint		*jBody;
  jsize		jNumTasks;
  jsize		jNumArgs;

  char		**cArgv;

  int		i;
  jstring	*jArgvElements;

	/* Convert jstrings to char * */

  cTask = (char *) getJavaString(jTask);
  cWhere = (char *) getJavaString(jWhere);

  if (_debug)
    fprintf(stdout, "jPVM.spawn(%s, *, %d, %s, *)\n",
	cTask, jFlag, cWhere);

	/* get the size and body of the int array */

  jNumTasks = (*env)->GetArrayLength(env, jTids);
  jBody = (*env)->GetIntArrayElements(env, jTids, 0);

	/* allocate space for C arrays and copy Argv into it */

  jNumArgs = (*env)->GetArrayLength(env, jArgv);
  cArgv = (char **) calloc( jNumArgs+1, sizeof(char *) );
  jArgvElements = (jstring *) calloc(jNumArgs, sizeof(jstring));

  for (i=0; i<jNumArgs; i++)
  {
      jArgvElements[i] = (jstring) (*env)->GetObjectArrayElement(env, jArgv, i);
      cArgv[i] = (char *) getJavaString(jArgvElements[i]);
  }

  cArgv[i] = NULL;

	/* Call the PVM Routine */

  jInfo = (jint) pvm_spawn( cTask, (char **) cArgv, (int) jFlag, cWhere,
                           (int) jNumTasks, (int *) jBody);

	/* Release Allocated Memory */

  (*env)->ReleaseIntArrayElements(env, jTids, jBody, 0);

  releaseJavaString(jTask, cTask);
  releaseJavaString(jWhere, cWhere);

  for(i=0; i<jNumArgs; i++)
      releaseJavaString(jArgvElements[i], cArgv[i]);

  free(jArgvElements);
  free(cArgv);

  return jInfo;
}

/* *******************************
 * jPVM.start_pvmd()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_start_pvmd(
	JNIEnv	 			*env, 
	jobject				obj,
	jint 				argc, 
	jobjectArray 			jArgv, 
	jint 				block)
{
  jint		jInfo;
  int		i;
  char		**cArgv;
  jstring	*jArgvElements;

  if (_debug)
    fprintf(stdout, "jPVM.start_pvmd(%d)\n", argc);

  if (argc == 0)
  {
    printf("jPVM:  no args to start_pvmd\n");
    jInfo = (jint) pvm_start_pvmd(0, NULL, (int) block);
  }
  else
  {
	/* allocate space for array of string arguments */

    cArgv = (char **) calloc(argc, sizeof(char *));
    jArgvElements = (jstring *) calloc (argc, sizeof(jstring *));
  
	/* copy arguments from Java string array */

    for (i=0; i<argc; i++)
    {
      jArgvElements[i] = (jstring) (*env)->GetObjectArrayElement(env, jArgv, i);
      cArgv[i] = (char *) getJavaString(jArgvElements[i]);
    }

	/* Call the PVM Routine */

    jInfo = (jint) pvm_start_pvmd( (int) argc, (char **) cArgv, (int) block);

	/* free allocated memory */

    for (i=0; i<argc; i++)
      releaseJavaString(jArgvElements[i], cArgv[i]);
    free(cArgv);

  }

  return(jInfo);
}

/* *******************************
 * jPVM.tidtohost()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_tidtohost(
	JNIEnv	 			*env,
	jobject				obj,
	jint				tid)
{
  if (_debug)
    fprintf(stdout, "jPVM.tidtohost(%d)\n", tid);

  return (jint) pvm_tidtohost( (int) tid);
}

/* *******************************
 * jPVM.upkbyte()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_trecv(
	JNIEnv	 			*env,
	jobject				obj,
	jint				tid,
	jint				msgtag,
	jint				sec,
	jint				usec)
{
  struct timeval tmout;

  if (_debug)
    fprintf(stdout, "jPVM.trecv(%d, %d, %d, %d)\n", 
		tid, msgtag, sec, usec);

	/* create the C struct for representing time */

  tmout.tv_sec = (int) sec;
  tmout.tv_usec = (int) usec;

	/* actual PVM call */

  return (jint) pvm_trecv((int) tid, (int) msgtag, &tmout);
}

/* ********************************
 * jPVM.upkbyte()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_upkbyte(
	JNIEnv	 			*env,
	jobject				obj,
	jbyteArray			jByteArray,
	jint				jNumItems,
	jint				jStride)
{
  jbyte		*jBody;
  char		*cBuf;
  jint		jInfo;
  jsize		jArrayLen;

  if (_debug)
    fprintf(stdout, "jPVM.upkbyte(*, %d, %d)\n", jNumItems, jStride);

	/* Error check to see if there are enough to hold the items */

  jArrayLen = (*env)->GetArrayLength(env, jByteArray);

  if (jArrayLen < jNumItems * jStride)
    return PvmBadParam;

	/* Get a pointer to the Java array */

  jBody = (*env)->GetByteArrayElements(env, jByteArray, 0);

	/* Call the routine */

  jInfo = (jint) pvm_upkbyte(jBody, (int) jNumItems, (int) jStride);

	/* Release VM memory */

  (*env)->ReleaseByteArrayElements(env, jByteArray, jBody, 0);


  return jInfo;
}


#if 0

/* *******************************
 * jPVM.upkcplx()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_upkcplx(
	JNIEnv	 			*env,
	jobject				obj,
	jfloatArray 			buf,
	jint				nitem,
	jint				stride)
{
  if (_debug)
    fprintf(stdout, "jPVM.upkcplx(*, %d, %d)\n", nitem, stride);


  return( (long) pvm_upkcplx(unhand(buf)->body, (int) nitem, (int) stride));
}

/* *******************************
 * jPVM.upkdclpx()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_upkdcplx(
	JNIEnv	 			*env,
	jobject				obj,
	jlongArray 			buf,
	jint				nitem,
	jint				stride)
{
  if (_debug)
    fprintf(stdout, "jPVM.upkdcplx(*, %d, %d)\n", nitem, stride);


  return( (long) pvm_upkdcplx(unhand(buf)->body, (int) nitem, (int) stride));
}

#endif

/* *******************************
 * jPVM.upkdouble()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_upkdouble(
	JNIEnv	 			*env,
	jobject				obj,
	jdoubleArray 			jDoubleArray,
	jint				jNumItems,
	jint				jStride)
{
  jdouble	*jBody;
  jint		jInfo;
  jsize		jArrayLen;

  if (_debug)
    fprintf(stdout, "jPVM.upkdouble(*, %d, %d)\n", jNumItems, jStride);

	/* Error check to see if there are enough to hold the items */

  jArrayLen = (*env)->GetArrayLength(env, jDoubleArray);
  if (jArrayLen < jNumItems * jStride )
    return PvmBadParam;

	/* Get a pointer to the Java array */

  jBody = (*env)->GetDoubleArrayElements(env, jDoubleArray, 0);

	/* Call the routine */

  jInfo = (jint) pvm_upkdouble(jBody, (int) jNumItems, (int) jStride);

	/* Release VM memory */

  (*env)->ReleaseDoubleArrayElements(env, jDoubleArray, jBody, 0);

  return jInfo;
}


/* *******************************
 * jPVM.upkfloat()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_upkfloat(
	JNIEnv	 			*env,
	jobject				obj,
	jfloatArray 			jFloatArray,
	jint				jNumItems,
	jint				jStride)
{
  jfloat	*jBody;
  jint		jInfo;
  int		jArrayLen, i;

  if (_debug)
    fprintf(stdout, "jPVM.upkfloat(*, %d, %d)\n", 
		jNumItems, jStride);

	/* Error check to see if there are enough to hold the items */

  jArrayLen = (*env)->GetArrayLength(env, jFloatArray);
  if (jArrayLen < jNumItems * jStride)
    return PvmBadParam;

	/* Get a pointer to the Java array */

  jBody = (*env)->GetFloatArrayElements(env, jFloatArray, 0);

	/* Call the routine */

  jInfo = (jint) pvm_upkfloat(jBody, (int) jNumItems, (int) jStride);

	/* Release VM memory */

  (*env)->ReleaseFloatArrayElements(env, jFloatArray, jBody, 0);

  return jInfo;

}

/* *******************************
 * jPVM.upkint()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_upkint(
	JNIEnv	 			*env,
	jobject				obj,
	jintArray			jIntArray,
	jint				jNumItems,
	jint				jStride)
{
  jint	*jBody;
  jint		jInfo;
  int		jArrayLen, i;

  if (_debug)
    fprintf(stdout, "jPVM.upkint(*, %d, %d)\n", 
		jNumItems, jStride);

	/* Error check to see if there are enough to hold the items */

  jArrayLen = (*env)->GetArrayLength(env, jIntArray);
  if (jArrayLen < jNumItems * jStride)
    return PvmBadParam;

	/* Get a pointer to the Java array */

  jBody = (*env)->GetIntArrayElements(env, jIntArray, 0);

	/* Call the routine */

  jInfo = (jint) pvm_upkint((int *) jBody, (int) jNumItems, (int) jStride);

	/* Release VM memory */

  (*env)->ReleaseIntArrayElements(env, jIntArray, jBody, 0);

  return jInfo;
}


/* *******************************
 * jPVM.upklong()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_upklong(
	JNIEnv	 			*env,
	jobject				obj,
	jlongArray 			jLongArray,
	jint				jNumItems,
	jint				jStride)
{
  jlong         *jBody;
  jint          jInfo;
  int           size;

  if (_debug)
    fprintf(stdout, "jPVM.upklong (*, %d, %d)\n", jNumItems, jStride);

        /* Error check to see if there are enough to hold the items */

  size = (*env)->GetArrayLength(env, jLongArray);
  if (size < (int) (jNumItems * jStride) )
    return PvmBadParam;

        /* Get a pointer to the Java array */

  jBody = (*env)->GetLongArrayElements(env, jLongArray, 0);

        /* Call the routine- use double */

  jInfo = (jint) pvm_upkdouble((double *) jBody, (int) jNumItems, 
		(int) jStride);

        /* Release VM memory */

  (*env)->ReleaseLongArrayElements(env, jLongArray, jBody, 0);

  return jInfo;

}


/* *******************************
 * jPVM.upkshort ()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_upkshort(
	JNIEnv	 			*env,
	jobject				obj,
	jshortArray 			jShortArray,
	jint				jNumItems,
	jint				jStride)
{
  jshort	*jBody;
  jint		jInfo;
  int		ArrayLength;

  if (_debug)
    fprintf(stdout, "jPVM.upkshort(*, %d, %d)\n", jNumItems, jStride);

	/* Error check to see if there are enough to hold the items */

  ArrayLength = (*env)->GetArrayLength(env, jShortArray);
  if (ArrayLength < (int) (jNumItems * jStride) )
    return PvmBadParam;

	/* Get a pointer to the Java array */

  jBody = (*env)->GetShortArrayElements(env, jShortArray, 0);

	/* Call the routine */

  jInfo = (jint) pvm_upkshort(jBody, (int) jNumItems, (int) jStride);

	/* Release VM memory */

  (*env)->ReleaseShortArrayElements(env, jShortArray, jBody, 0);

  return jInfo;

}



/* *******************************
 * jPVM.upkstr()
 * ********************************/

JNIEXPORT jint JNICALL
  Java_jPVM_upkstr(
	JNIEnv	 			*env,
	jobject				obj,
	jcharArray			jCharArray)
{
  jint		jInfo;
  jsize		jArrayLen;
  jchar		*jBody;
  char		*sp;
  int		i;
  int		cSize;

  if (_debug)
    fprintf(stdout, "jPVM.upkstr(*)\n");

	/*
	 * Allocate C array of characters equal to length of
	 * uniStr array of jchars.
	 */

  jArrayLen = (*env)->GetArrayLength(env, jCharArray);

  sp = (char *) calloc(jArrayLen, sizeof(char));

        /* Actual PVM routine */
 
  jInfo = (jint) pvm_upkstr(sp);
  
  cSize = strlen(sp);

  if (_debug)
    fprintf(stdout, 
	"jPVM.pvm_upkstr:  C string = %s  length = %d\n", sp, cSize);

	/*
	 * Get body of Java character array
	 * and copy C character array into it.
	 * But only copy up to null terminator 
	 * C character array.  Don't use strcpy()
	 * because we can't be sure how jChar is
	 * represented on different platforms.
	 */

  jBody = (*env)->GetCharArrayElements(env, jCharArray, 0);

  for (i=0; i<jArrayLen; i++)
    jBody[i] = sp[i];

	/*
	 * Release VM allocated memory (array body)
	 * and locally allocated memory.
	 */

  (*env)->ReleaseCharArrayElements(env, jCharArray, jBody, 0);

  free(sp);

  return jInfo;
}
